/***************************************************************************

    file        : raceselectmenu.cpp
    created     : Sat Nov 16 09:36:29 CET 2002
    copyright   : (C) 2002 by Eric Espie
    email       : eric.espie@torcs.org   
    version     : $Id: raceselectmenu.cpp 7827 2021-12-17 00:57:48Z iobyte $                                  

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/** @file   
    		Race selection menu
    @author	<a href=mailto:eric.espie@torcs.org>Eric Espie</a>
    @version	$Id: raceselectmenu.cpp 7827 2021-12-17 00:57:48Z iobyte $
*/

#include <map>
#include <algorithm>
#include <string>

#include <tgfclient.h>
#include <raceman.h>
#include <racemanagers.h>
#include <race.h>
#include <cars.h>

#include "legacymenu.h"
#include "racescreens.h"

class GfCars;
class GfDrivers;
class GfRaces;

// The Race Select menu.
void *RmRaceSelectMenuHandle = NULL;

// Menu control Ids
static int PracticeButtonId;
static int Challenge1ButtonId;
static int Challenge2ButtonId;
static int Endurance1ButtonId;
static int ChampionshipButtonId;
static int QuickraceButtonId;

static int NextButtonId;

const char* rmCategory = "";
int         rmRanking[];
int         rmIndex = 0;
int         rmSkilling = 0;
int         rmPosition = 0;

const GfCar*    _pCar;
std::string     _pCategory;

const char* rmrank[] = { "ranking0", "ranking1", "ranking2", "ranking3",
                        "ranking4", "ranking5", "ranking6", "ranking7",
                        "ranking8", "ranking9", "ranking10", "ranking11",
                        "ranking12", "ranking13", "ranking14", "ranking15",
                        "ranking16", "ranking17", "ranking18", "ranking19" };

std::string  _pCat[] = { "11GPA-", "SC-C-", "12GPA-", "SC-B-", "01GPA-", "SC-A-",
                         "LS2-", "RS-", "LS1-", "HC-", "LP2-", "36GP-", "LP1-",
                         "67GP-", "91GP-", "NC-", "01GP-", "PR-" };


/* Called when the menu is activated */
static void
rmOnActivate(void * /* dummy */)
{
	GfLogTrace("Entering Race Mode Select menu\n");

    LmRaceEngine().reset();
}

/* Exit from Race engine */
static void
rmOnRaceSelectShutdown(void *prevMenu)
{
    GfuiScreenActivate(prevMenu);
	
    LmRaceEngine().cleanup();
	
	LegacyMenu::self().shutdownGraphics(/*bUnloadModule=*/true);
}

static void
rmOnSelectRaceManPractice(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = _pCategory+"practice";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);
    GfLogInfo("Race Practice = %s\n", race.c_str());

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static void
rmOnSelectRaceManQuickrace(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = "quickrace";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static void
rmOnSelectRaceManChallenge1(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = _pCategory+"challenge1";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static void
rmOnSelectRaceManChallenge2(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = _pCategory+"challenge2";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static void
rmOnSelectRaceManEndurance1(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = _pCategory+"endurance1";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static void
rmOnSelectRaceManChampionship1(void *pvRaceManTypeIndex)
{
    // Get the race managers with the given type
    std::string race = _pCategory+"championship1";
    GfRaceManager* pSelRaceMan = GfRaceManagers::self()->getRaceManager(race);

    if (pSelRaceMan)
    {
        // Give the selected race manager to the race engine.
        LmRaceEngine().selectRaceman(pSelRaceMan);

        // Start the race configuration menus sequence.
        LmRaceEngine().configureRace(/* bInteractive */ true);
    }
    else
    {
        GfLogError("No such race manager (type '%s')\n", race.c_str());
    }
}

static
void rmRaceCategoryInit()
{
    GfLogInfo("Prefix category Race : \n");
    if (strcmp(rmCategory, "2011 GP American") == 0)
    {
        rmIndex    = 0;
        _pCategory = _pCat[rmIndex];

    }
    else if(strcmp(rmCategory, "Supercars C") == 0)
    {
        rmIndex    = 1;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "2012 GP American") == 0)
    {
        rmIndex    = 2;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Supercars B") == 0)
    {
        rmIndex    = 3;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "2001 GP American") == 0)
    {
        rmIndex    = 4;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Supercars A") == 0)
    {
        rmIndex    = 5;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Long Day Series GT2") == 0)
    {
        rmIndex    = 6;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Rally Series") == 0)
    {
        rmIndex    = 7;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Long Day Series GT1") == 0)
    {
        rmIndex    = 8;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Hypercars") == 0)
    {
        rmIndex    = 9;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Long Day Series P2") == 0)
    {
        rmIndex    = 10;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "1936 Grand Prix") == 0)
    {
        rmIndex    = 11;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Long Day Series P1") == 0)
    {
        rmIndex    = 12;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "1967 Grand Prix") == 0)
    {
        rmIndex    = 13;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "1991 Grand Prix") == 0)
    {
        rmIndex    = 14;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Nascar") == 0)
    {
        rmIndex    = 15;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "2001 Grand Prix") == 0)
    {
        rmIndex    = 16;
        _pCategory = _pCat[rmIndex];
    }
    else if (strcmp(rmCategory, "Prototype") == 0)
    {
        rmIndex    = 17;
        _pCategory = _pCat[rmIndex];
    }

    GfLogInfo("Prefix category Race : %s - Ranking index = %i\n", _pCategory.c_str(), rmIndex);
}

static void
rmRaceSelectNextPage(void *prevMenu)
{
    GfuiScreenActivate(RmRaceSelectInit2(RmRaceSelectMenuHandle));
}

/* Initialize the single player menu */
void *
RmRaceSelectInit(void *prevMenu)
{
    if (RmRaceSelectMenuHandle)
    {
        RmRaceSelectMenuHandle = NULL;
    }

    std::string strPlayersFileName(GfLocalDir());
    strPlayersFileName += "drivers/human/human.xml";
    _pCategory = "";

    void *hparmHuman = GfParmReadFileLocal(strPlayersFileName, GFPARM_RMODE_STD | GFPARM_RMODE_REREAD);

    if (!hparmHuman)
        hparmHuman = GfParmReadFile(strPlayersFileName, GFPARM_RMODE_STD | GFPARM_RMODE_REREAD);

    if (!hparmHuman)
    {
        GfLogError("No usable human driver (human.xml not found or not readable)\n");
    }

    if (hparmHuman)
    {
        char sstring[128];
        //const char *str;

        GfLogInfo("Loadind race select init human\n");

        snprintf(sstring, sizeof(sstring), "%s/%s/%d", "Robots", "index", 1);
        for (int i = 0; i < 18; i++)
        {
            rmRanking[i] = GfParmGetNum(hparmHuman, sstring, rmrank[i], (char*)NULL, 0);
            GfLogInfo("_rank %i          = %i\n", i, rmRanking[i]);
        }

        const char* pszCarId = GfParmGetStr(hparmHuman, sstring, "car name", "");
        std::string car = GfCars::self()->getCar(pszCarId)->getCategoryName();
        rmCategory = car.c_str();

        GfLogInfo("Car              = %s\n", rmCategory);

        rmRaceCategoryInit();

        GfParmReleaseHandle(hparmHuman);
    }

	// Ask the RaceEngine what types of races should be allowed here
    //bool SupportsHumanDrivers = LmRaceEngine().supportsHumanDrivers();

    // Create screen, load menu XML descriptor and create static controls.
    RmRaceSelectMenuHandle = GfuiScreenCreate((float*)NULL, 
											NULL, rmOnActivate, 
											NULL, (tfuiCallback)NULL, 
											1);
#if defined(CLIENT_SERVER)
    void *hMenuXMLDesc = GfuiMenuLoad("csraceselectmenu.xml");
#else
	void *hMenuXMLDesc = GfuiMenuLoad("raceselectmenu.xml");
#endif
    GfuiMenuCreateStaticControls(RmRaceSelectMenuHandle, hMenuXMLDesc);

    // Create the raceman type buttons and sub-type combo-boxes (if any).
    PracticeButtonId =
        GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "PracticeButton",
                            NULL, rmOnSelectRaceManPractice);
    Challenge1ButtonId = GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "Challenge1Button",
                                                        NULL, rmOnSelectRaceManChallenge1);
    Challenge2ButtonId = GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "Challenge2Button",
                                                        NULL, rmOnSelectRaceManChallenge2);
    Endurance1ButtonId = GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "Endurance1Button",
                                                        NULL, rmOnSelectRaceManEndurance1);
    ChampionshipButtonId = GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "Championship1Button",
                                                       NULL, rmOnSelectRaceManChampionship1);
    QuickraceButtonId = GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "QuickraceButton",
                                                NULL, rmOnSelectRaceManQuickrace);

    if(rmRanking[rmIndex] < 1 )
        GfuiEnable(RmRaceSelectMenuHandle, Challenge1ButtonId, GFUI_DISABLE);

    if(rmRanking[rmIndex] < 2 )
        GfuiEnable(RmRaceSelectMenuHandle, Challenge2ButtonId, GFUI_DISABLE);

    if(rmRanking[rmIndex] < 3 )
        GfuiEnable(RmRaceSelectMenuHandle, Endurance1ButtonId, GFUI_DISABLE);

    if(rmRanking[rmIndex] < 4 )
        GfuiEnable(RmRaceSelectMenuHandle, ChampionshipButtonId, GFUI_DISABLE);

    if (rmRanking[16] < 0 && rmRanking[17] < 0)
        GfuiVisibilitySet(RmRaceSelectMenuHandle, QuickraceButtonId, GFUI_INVISIBLE);
	
    // Create Back button
    GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "BackButton",
                                prevMenu, rmOnRaceSelectShutdown);

    // Create Next button
    NextButtonId =
            GfuiMenuCreateButtonControl(RmRaceSelectMenuHandle, hMenuXMLDesc, "NextButton",
                                NULL, rmRaceSelectNextPage);

    // Close menu XML descriptor.
    GfParmReleaseHandle(hMenuXMLDesc);
    
    // Register keyboard shortcuts.
    GfuiMenuDefaultKeysAdd(RmRaceSelectMenuHandle);
    GfuiAddKey(RmRaceSelectMenuHandle, GFUIK_ESCAPE, "Back To Main Menu",
               prevMenu, rmOnRaceSelectShutdown, NULL);

    // Give the race engine the menu to come back to.
    LmRaceEngine().initializeState(RmRaceSelectMenuHandle);

    return RmRaceSelectMenuHandle;
}
